/* mbiff.c - pseudo-biff that displays a one-line "scan" listing */

#include "../h/mh.h"
#include "../h/formatsbr.h"
#include "../h/scansbr.h"
#include "../zotnet/tws.h"
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

static struct swit switches[] = {
#define	CHECKSW	0
	"checkint", 0,

#define	FORMSW	1
	"form formatfile", 0,
#define	FMTSW	2
	"format string", 5,

#define	WIDSW	3
	"width columns", 0,

#define	BIFFSW	4
	"biff", 0,
#define	NBIFFSW	5
	"nobiff", 0,

#define	BELLSW	6
	"bell", 0,
#define	NBELLSW	7
	"nobell", 0,

#define	MARKSW	8
	"mark", 0,
#define	NMARKSW	9
	"nomark", 0,

#define	HELPSW	10
	"help", 4,

	NULL, NULL
};

#ifndef	FD_SETSIZE
#define	FD_SET(n, p)	((p)->fds_bits[0] |= (1<<(n)))
#define	FD_CLR(n, p)	((p)->fds_bits[0] &= ~(1<<(n)))
#define	FD_ISSET(n, p)	((p)->fds_bits[0] & (1<<(n)))
#define FD_ZERO(p)	((p)->fds_bits[0] = 0)
#endif

extern int errno;
int biff_fd = -1;
off_t last_size;
char *user;

/*
 * wait for a biff message or for maxwait seconds, whichever comes
 * first.  If maxwait is zero, wait only for biff msg.  If biff_fd
 * is zero, wait maxwait seconds.
 */
void
waitformsg(maxwait)
	int maxwait;
{
	fd_set fds;
	struct timeval wait_time;
	int i;
	int gotoffset = 0;
	char *cp;
	char msgbuf[128];

	if (biff_fd < 0) {
		sleep(maxwait);
		return;
	}
again:
	FD_ZERO(&fds);
	FD_SET(biff_fd, &fds);
	if (maxwait) {
		wait_time.tv_sec = maxwait;
		wait_time.tv_usec = 0;
		i = select(biff_fd + 1, &fds, 0, 0, &wait_time);
	} else
		i = select(biff_fd + 1, &fds, 0, 0, 0);
	if (i != 1 || !FD_ISSET(biff_fd, &fds))
		return;

	/* read and discard biff msg(s) */
	wait_time.tv_sec = 0;
	wait_time.tv_usec = 500000;
	while (i == 1 && FD_ISSET(biff_fd, &fds)) {
		if (recv(biff_fd, msgbuf, sizeof(msgbuf), 0) <= 0)
			break;
		if ((cp = index(msgbuf, '@')) == 0)
			goto again;
		*cp = 0;
		if (strcmp(msgbuf, user))
			goto again;
		if (gotoffset == 0) {
			last_size = atoi(cp + 1);
			gotoffset = 1;
		}
		i = select(biff_fd + 1, &fds, 0, 0, &wait_time);
	}
}

/* ARGSUSED */
main(argc, argv)
	int argc;
	char *argv[];
{
	int bellflag = 1,
	    markflag = 1,
	    needmark = 0,
	    checkint = 30,
	    width = 0,
	    usebiff = 0,
	    state, nmsgs;
	char *cp, *file = NULL,
	    *form = NULL, *format = NULL, buf[100],
	    **ap, **argp, *nfs, *arguments[MAXARGS];
	FILE *in;

	invo_name = r1bindex(argv[0], '/');
	if ((cp = m_find(invo_name)) != NULL) {
		ap = brkstring(cp = getcpy(cp), " ", "\n");
		ap = copyip(ap, arguments);
	} else
		ap = arguments;
	(void) copyip(argv + 1, ap);
	argp = arguments;

	while (cp = *argp++) {
		if (*cp == '-')
			switch (smatch(++cp, switches)) {

			case AMBIGSW:
				ambigsw(cp, switches);
				done(1);
			case UNKWNSW:
				adios(NULLCP, "-%s unknown", cp);
			case HELPSW:
				(void) sprintf(buf,
				    "%s [+folder] [msgs] [switches]",
				    invo_name);
				help(buf, switches);
				done(1);

			case FORMSW:
				if (!(form = *argp++) || *form == '-')
					adios(NULLCP,
					    "missing argument to %s", argp[-2]);
				format = NULL;
				continue;
			case FMTSW:
				if (!(format = *argp++) || *format == '-')
					adios(NULLCP,
					    "missing argument to %s", argp[-2]);
				form = NULL;
				continue;

			case CHECKSW:
				if (!(cp = *argp++) || *cp == '-')
					adios(NULLCP,
					    "missing argument to %s", argp[-2]);
				checkint = atoi(cp);
				continue;

			case WIDSW:
				if (!(cp = *argp++) || *cp == '-')
					adios(NULLCP,
					    "missing argument to %s", argp[-2]);
				width = atoi(cp);
				continue;

			case BIFFSW:
				usebiff = 1;
				checkint = 0;
				continue;
			case NBIFFSW:
				usebiff = 0;
				continue;

			case BELLSW:
				bellflag = 1;
				continue;
			case NBELLSW:
				bellflag = 0;
				continue;

			case MARKSW:
				markflag = 1;
				continue;
			case NMARKSW:
				markflag = 0;
				continue;
			}
		if (file)
			adios(NULLCP, "only one file at a time!");
		else
			file = path(cp, TFILE);
	}
	if (usebiff) {
		struct servent *sp = getservbyname("biff", "udp");
		struct sockaddr_in sa;

		if (! sp)
			adios(NULLCP, "can't find `biff' port number.");

		biff_fd = socket(AF_INET, SOCK_DGRAM, 0);
		if (biff_fd < 0) {
			perror("mbiff: socket");
			exit (1);
		}
		bzero(&sa, sizeof(sa));
		sa.sin_family = AF_INET;
		sa.sin_port = sp->s_port;
		if (bind(biff_fd, &sa, sizeof(sa)) < 0) {
			perror("mbiff: bind");
			exit (1);
		}
	} else if (checkint <= 0)
		checkint = 30;

	setuid(getuid());

	user = getenv("USER");
	nfs = new_fs(form, format, FORMAT);
	if (file == NULL) {
		(void)sprintf(buf, "/usr/spool/mail/%s", user);
		file = buf;
	}
	in = fopen(file, "r");
	if (in == NULL)
		adios(NULLCP, "can't open %s (%d)", file, errno);

	nmsgs = 1;
	last_size = 0;
	for ( ; ; waitformsg(checkint)) {
		struct stat st;

		/* sleep until there's something in the file */
		(void) fstat(fileno(in), &st);
		if (st.st_size == 0) {
			last_size = 0;
			continue;
		}
		/* sleep until the file changes size */
		if (st.st_size == last_size)
			continue;

		/*
		 * if the new size is smaller than the old, go back to
		 * the beginning of the file.
		 */
		if (st.st_size < last_size)
			last_size = 0;

		(void)fseek(in, last_size, 0);

		/* if we're just starting, figure out the mbox type */
		if (last_size == 0) {
			m_unknown(in);
			if (markflag && needmark) {
				putchar('\n');
				needmark = 0;
			}
		}
		while (1) {
			state = scan(in, nmsgs, -1, nfs, width, 0, 0, 0L, 1, 0);
			if (state != SCNMSG && state != SCNENC)
				break;
			needmark = 1;
			++nmsgs;
		}
		/* remember where we ended up */
		last_size = ftell(in);
		if (bellflag)
			putchar('');
		fflush(stdout);
	}
	/* NOTREACHED */
}
